//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/gensig/main.c,v $
// $Id: main.c,v 1.24 2005/10/29 10:44:04 ah Exp $ 
//
// Description: print information about specified wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: main.c,v 1.24 2005/10/29 10:44:04 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>
#include <iostream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Config/Types.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/Math/Dither.h>
#include <AH/Error/Class.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include "Signals.h"
#include "Util.h"

// --- local definitions ---


// --- local constants ---

static const char* cmdLineOptions = "C:D:b:c:dhmN:n:r:s:t:v";
static const char* version = "1.2";

// --- local data ---

static unsigned int debon = 0;
static unsigned int channels = 1;
static unsigned int seconds = 60;       // 1 minute
static unsigned int samples = 0;
static unsigned int samplerate = 44100;
static unsigned int samplebits = 16;
static double norm_ampl = 0;
static MathDither::Type dither = MathDither::None;
static std::string out_file;

// --- local classes ---

// --- local functions ---

static void usage(bool show_examples = false);


static void usage(bool show_examples)
{
    cerr << "Usage: gensig [-" << cmdLineOptions << "] outputfile\n"
	 << "       generate a WAV file with one or more test signals.\n"
	 << "  options:\n"
	 << "    b <bits>    - number of sample bits (default = 16)\n"
	 << "    c <num>     - number of channels\n"
	 << "                  mono=1, stereo=2 (default=" << channels << ")\n"
	 << "    C <num>     - generate following signals for this channel:\n"
	 << "                    0 (all channels), 1 (left), 2 (right), .. (default=0)\n"
	 << "                  this option can be given before each -s\n"
	 << "    d           - increase debug output (may be given more than once)\n"
	 << "    D <typ>     - add dither\n"
	 << "                  typ=r: rectangulare PDF\n"
	 << "                  typ=t: triangle PDF\n"
	 << "                  typ=n: no dither, round to next Bit (default)\n"
	 << "    h           - print this help message\n"
	 << "    m           - print MLS table\n"
	 << "    N <samples> - number of samples (or use -t)\n"
	 << "    n <ampl>    - normalize amplitude (float, or in dB)\n"
	 << "    r <rate>    - sample rate in Samples/s (Hz, default = 44100)\n"
	 << "    s <cmd>     - specify signal (may be given more than once);\n"
	 << "                  <cmd> is one of the following sequences:\n"
	 << Signals::help("                  ")
	 << "    t <sec>     - time in seconds (default = 60)\n"
	 << "    v           - print version\n"
	 << "  output file   - WAV file\n" << endl;
    if (!show_examples) return;

    cerr << "see the following examples:\n"
	 << "1. gensig -s\"sinus, 1000, 0 dB\" sin1000.wav\n"
	 << "   This generates the WAV 'sin1000.wav' file with a 1 kHz sine wave at full\n"
	 << "   amplitude. The WAV file is mono, 44100 Hz samplerate, 16 bits samples\n"
	 << "   and 1 minute long.\n"
	 << "2. gensig -t 10 -s\"sinus, 1000, -6 dB\" -s\"sinus, 10000, -20 dB\" sinmix.wav\n"
	 << "   This generates the WAV file 'sinmix.wav' with a 1 kHz sine wave at -6 dB\n"
	 << "   and a 10 kHz sine wave at -20 dB. It is 10 seconds long.\n"
	 << "3. gensig -t 5 -c 2 -r 96000 -b 24 -s\"square, 315, -10 dB\" square.wav\n"
	 << "   This generates the WAV file 'square.wav' with a 315 Hz square wave at\n"
	 << "    -6 dB. It is stereo, 96 kHz samplerate, 24 bits samples and 5 seconds\n"
	 << "   long.\n"
	 << "4. gensig -n \"0 dB\" -s\"sinus, 14000, 1\" -s\"sinus, 15000, 1\" imf.wav\n"
	 << "   This generates the WAV file 'imf.wav' with a 14 kHz and a 15 kHz sine\n"
	 << "    wave of identical amplitude. The sinus mix is normalized to full\n"
	 << "    amplitude (0 dB). It is 1 minute long.\n"
	 << "5. gensig -c 2 -C 1 -s\"saw, 100, -10 dB\" -C 2 -s\"saw, 50, -15 dB\" lr.wav\n"
	 << "   This generates the WAV file 'lr.wav' with a 100 Hz square wave at -10 dB\n" 
	 << "   on the left channel and a 50 Hz square wave at -15 dB on the right channel.\n"
	 << endl;
} // end of usage()



static void print_version()
{
    cerr << "gensig Vs " << version << " " 
	 << "(c) 2002 Andreas Hnnebeck\n"
	 << "sweep algorithm (c) 2002 Patrick Piecha\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()



static void process(Signals* sig[])
{
    // get number of samples to create:
    unsigned int num;
    if (samples) num = samples;
    else         num = samplerate * seconds;

    // create empty buffer for signal (double):
    double** data = new double*[channels];
    if (!data) throw ErrorSys("process", ENOMEM);
    for (unsigned int j = 0; j < channels; ++j)
    {
	data[j] = new double[num];
	if (!data[j]) throw ErrorSys("process", ENOMEM);
	memset(data[j], 0, num * sizeof(*data[j]));
    }

    // create the desired signals in the buffer:
    if (sig)
    {
	// first the signals which go into each channel:
	if (sig[0])
	{
	    sig[0]->calculate(data[0], num);
	    for (unsigned int j = 1; j < channels; ++j)
		memcpy(data[j], data[0], num * sizeof(*data[j]));
	}

	// now the signals which go into separate channels:
	for (unsigned int j = 0; j < channels; ++j)
	{
	    if (sig[j+1])
		sig[j+1]->calculate(data[j], num);
	    if (0 != norm_ampl) Util::normalize(data[j], num, norm_ampl);
	}
    }
    MathDither dith(dither);

    // create WAV file:
    IORiffWaveFile wave(out_file.c_str(), channels, samplebits, samplerate,
			num);
    
    // get size of data for desired samples
    const uint32 snum = 65536; // number of samples to write
    const uint32 bufsize = wave.getDataBufSize(snum);

    uint32 i = 0;
    int ret = 0;
    do
    {
	unsigned char sbuf[bufsize];
	memset(sbuf, 0, bufsize);

	// set size of data to read on this turn:
	uint32 rnum = snum;
	uint32 rbufsize = bufsize;
	if (rnum > num)
	{
	    rnum = num;
	    rbufsize = wave.getDataBufSize(num);
	}

	// write double data (dithered) into output buffer
	for (uint32 k = 0; k < rnum; ++k)
	{
	    for (unsigned int j = 0; j < channels; ++j)
	    {
		int32 val = (int32)(floor(data[j][i] + 0.5 + 
					  dith.getSignal()));
		if (debon > 3)
		    cerr << "write_file: val = " << val << endl;
		wave.setBufSample(sbuf, k, j+1, val);
	    }
	    ++i;
	}

	// write output buffer:
	if (debon > 2) 
	    cerr << "write_file: write " << rnum << " points" << endl;
	ret = wave.writeRaw(sbuf, rbufsize);
	if (ret <= 0) break;
	if ((uint32)ret < rbufsize)
	    cerr << "Warning: could not write " << (rbufsize - ret) 
		 << " bytes to WAV file" << endl;  
	num -= rnum;
    } while (num > 0);

    for (unsigned int j = 0; j < channels; ++j)
	delete[] data[j];
    delete[] data;

    wave.writeHeader();
}


// --- global functions ---


int main(int argc, char * const argv[])
{
    srandom(0);
    Signals** sigs = 0;

    static unsigned int errnum = 0;
    try
    {
	// loop through command line options:
	int option;
	unsigned int chan = 0;
	while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
	{
	    switch (option) 
	    {
	    case 'C': chan = atoi(optarg);
		      if (chan > channels) throw ErrorBasic(
		          "bad number of specified channel");
	              break;

	    case 'D': switch(optarg[0])
	              {
		      default:
		      case 'n': dither = MathDither::None; break;
		      case 'r': dither = MathDither::RectAngle; break;
		      case 't': dither = MathDither::Triangle; break;
		      }
	              break;

	    case 'b': if (sigs) throw ErrorBasic(
		          "argument -b must come before -s");
	              if (0 != norm_ampl) throw ErrorBasic(
		          "argument -b must come before -n");
	              samplebits = atoi(optarg);
		      if (samplebits <= 0) throw ErrorBasic(
			  "bad numer of sample bits");
		      break;
	    case 'c': channels = atoi(optarg); break;
	    case 'd': if (sigs) throw ErrorBasic(
		          "argument -d must come before -s");
	              ++debon; 
		      break;
	    case 'h': usage(true); exit(0);
	    case 'm': cerr << Signals::mlsTable("  "); exit(0);
            case 'N': if (sigs) throw ErrorBasic(
		          "argument -N must come before -s");
	              samples = atoi(optarg);
		      break;
	    case 'n': norm_ampl = Util::getAmplitude(optarg, samplebits);
		      break;
	    case 'r': if (sigs) throw ErrorBasic(
		          "argument -r must come before -s");
	              samplerate = atoi(optarg);
		      if (samplerate <= 0) throw ErrorBasic(
			  "bad samplerate");
		      break;
	    case 's': if (!sigs) 
	              {
		          sigs = new Signals*[channels+1];
			  if (!sigs) throw ErrorSys("main", ENOMEM);
			  for (unsigned int j = 0; j < channels+1; ++j)
			      sigs[j] = 0;
		      }
	              if (!sigs[chan])
		      {
			  sigs[chan] = new Signals(samplerate, samplebits, 
						   seconds, debon);
			  if (!sigs[chan]) throw ErrorSys("process", ENOMEM);
		      }
		      sigs[chan]->add(optarg);
		      break;
	    case 't': if (sigs) throw ErrorBasic(
		          "argument -t must come before -s");
	              seconds = atoi(optarg);
	              samples = 0; // override option '-N' 
		      break;
	    case 'v': print_version(); exit(0);
	    default : usage(); exit(1);
	    }
	}
	if (optind >= argc) 
	{
	    cerr << "Error: missing output file" << endl;
	    usage();
	    exit(1);
	}

	// loop through the remaining arguments: 
	vector<string> args;
	for (int i=optind; i < argc; i++)
	{
	    args.push_back(argv[i]);
	}

	switch (args.size())
	{
	case 0:  cerr << "Error: missing output file" << endl;
	         usage();
		 exit(1);
	case 1:  out_file = args[0];
	         break;
	default: cerr << "Error: too many arguments" << endl;
	         usage();
		 exit(1);
	}

	if (debon > 1)
	{
	    cerr << "- channels   = " << channels << "\n";
	    cerr << "- seconds    = " << seconds << "\n";
	    cerr << "- samples    = " << samples << "\n";
	    cerr << "- samplerate = " << samplerate << "\n";
	    cerr << "- samplebits = " << samplebits << "\n";
	    cerr << "- norm_ampl  = " << norm_ampl << "\n";
	    cerr << "- dither     = " << dither << "\n";
	    cerr << "- out_file   = " << out_file << "\n";
	}

	// process the files
	if (debon) cerr << "main: process the signals" << endl;
	process(sigs);
    }
    catch (ErrorClass& err)
    {
	++errnum;
	cerr << "Error: " << err.getMsg() << endl;
    }
    catch (exception& err)
    {
	++errnum;
	cerr << "Error: " << err.what() << endl;
    }
    
    if (sigs)
    {
	for (unsigned int j = 0; j < channels+1; ++j)
	    delete sigs[j];
	delete[] sigs;
    }

    exit (errnum ? 1 : 0);
    return (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---



