//---------------------------------------------------------------------------
// Copyright 1999 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/IO/AudioFile.h,v $
// $Id: AudioFile.h,v 1.2 2005/11/03 21:51:17 ah Exp $
//
// Description: 
//
//---------------------------------------------------------------------------

#ifndef AH_IO_AUDIO_FILE_H		// prevent multiple includes
#define AH_IO_AUDIO_FILE_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
#include <AH/Config/Types.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include <cstdio>
#include <cerrno>
#include <iostream>
#include <string>
#include <vector>

#include <sndfile.h>

// --- types and classes ---

BEGIN_NAMESPACE(AH)

//----------------------------------------------------------------------------------
// Policy classes
//----------------------------------------------------------------------------------

//----------------------------------------------------------------------------------
/** \defgroup Audio File 'Open' Policy classes */
//\@{

/// generic policy to open any audio file
/**
   IOAudioFileOpenPolicy uses a conversion program to convert an audio file of
   a format which cannot be handled by libsndfile, into a format that can be
   handled (e.g. into uncompressed WAV). Since this is a pure virtual class you must
   derive a subclass to be able to use it.
*/
class IOAudioFileOpenPolicy
{
 public:
    /// constructor
    IOAudioFileOpenPolicy
    (
	const std::string& ConverterName, ///< conversion program
	const std::string& Extension      ///< extension of audio file
    )
	: converterName(ConverterName)
    {
	extensions.push_back(Extension);
    }

    /// destructor
    virtual ~IOAudioFileOpenPolicy() {}

    /// open the file
    /**
       \return pointer to open file
       \return NULL in case of error
     */
    FILE* openFile
    (
	const std::string& filename, ///< name of file
	std::string& ConverterName   ///< output: set to name of conversion program used to open file
    ) const
    {
	ConverterName = converterName;
	return openFile(filename);
    }

    /// return true if given extension matches any of the extensions of the conversion program
    /**
       matchesExtension() compares the given extension with all supported extensions.
       \return true if given extension matches any of the supported extensions
       \return true if any of the supported extensions is empty
       \return false in all other cases
     */
    virtual bool matchesExtension(const std::string& extension) const
    {
	for (unsigned int i = 0; i < extensions.size(); ++i)
	{
	    if (extensions[i].empty() || 
		(0 == strcasecmp(extensions[i].c_str(), extension.c_str())) )
		return true;
	}
	return false;
    }

    /// return name of conversion program
    std::string getConverterName() const
	{ return converterName; }

 protected:
    virtual FILE* openFile(const std::string& filename) const = 0;
    std::string converterName;
    std::vector<std::string> extensions;
};


/// policy to open a known file format.
/** 
    IOAudioFileOpenPolicyFile opens the file directly without conversion.
 */
class IOAudioFileOpenPolicyFile: public IOAudioFileOpenPolicy
{
 public:
    /// default constructor
    IOAudioFileOpenPolicyFile()
	: IOAudioFileOpenPolicy("", "") {}

    /// open the file
    virtual FILE* openFile(const std::string& filename) const
    {
	if ((filename == "stdin") || (filename == "-"))
	    return stdin;
	return fopen(filename.c_str(), "r");
    }
};


/// policy to open an MP3 file by converting it into WAV format
/** 
    IOAudioFileOpenPolicyMP3 uses <i>mpg123</i> to open MP3 files as WAV file. 
 */
class IOAudioFileOpenPolicyMP3: public IOAudioFileOpenPolicy
{
 public:
    /// default constructor
    IOAudioFileOpenPolicyMP3()
	: IOAudioFileOpenPolicy("mpg123", "mp3") {}

 protected:
    /// open the file
    virtual FILE* openFile(const std::string& filename) const
    {
	// use mpg123 to convert MP3 file in WAV
	std::string cmd = converterName + " -q -w - \"" + filename + "\" 2> /dev/null";
	return popen(cmd.c_str(), "r");
    }
};

/// policy to open an OGG/Vorbis file by converting it into WAV format
/** 
    IOAudioFileOpenPolicyMP3 uses <i>oggdec</i> to open OGG/Vorbis files as WAV file. 
 */
class IOAudioFileOpenPolicyOGG: public IOAudioFileOpenPolicy
{
 public:
    /// default constructor
    IOAudioFileOpenPolicyOGG()
	: IOAudioFileOpenPolicy("oggdec", "ogg") {}

 protected:
    /// open the file
    virtual FILE* openFile(const std::string& filename) const
    {
	// use oggdec to convert ogg file in WAV
	std::string cmd = converterName + " -Q -o - \"" + filename + "\" 2> /dev/null";
	return popen(cmd.c_str(), "r");
    }
};

/// policy to open an MP+ file by converting it into WAV format
/** 
    IOAudioFileOpenPolicyMPP uses <i>mppdec</i> to open MP3+ files as WAV file. 
 */
class IOAudioFileOpenPolicyMPP: public IOAudioFileOpenPolicy
{
 public:
    /// default constructor
    IOAudioFileOpenPolicyMPP()
	: IOAudioFileOpenPolicy("mppdec", "mpc")
    {
	extensions.push_back("mp+");
	extensions.push_back("mpp");
    }

 protected:
    /// open the file
    virtual FILE* openFile(const std::string& filename) const
    {
	// use mppdec to convert Mpeg+ file in WAV
	std::string cmd = converterName + " --silent - - < \"" + filename + "\" 2> /dev/null";
	return popen(cmd.c_str(), "r");
    }
};

/// policy to open an LPAC file by converting it into WAV format
/** 
    IOAudioFileOpenPolicyLPAC uses <i>lpac</i> to open lossless compressed
    LPAC files as WAV file. 
 */
class IOAudioFileOpenPolicyLPAC: public IOAudioFileOpenPolicy
{
 public:
    /// default constructor
    IOAudioFileOpenPolicyLPAC()
	: IOAudioFileOpenPolicy("lpac", "pac") {}

 protected:
    /// open the file
    virtual FILE* openFile(const std::string& filename) const
    {
	// use lpac to convert pac file in WAV
        std::string cmd = converterName + " -x \"" + filename + "\" -  2> /dev/null";
	return popen(cmd.c_str(), "r");
    }
};


/// opens audio file
/** 
    IOAudioFileOpener bundles builtin and user defined policies to open
    audio files.
    \see IOAudioFileOpenPolicy
 */
class IOAudioFileOpener
{
 public:
    typedef std::vector<IOAudioFileOpenPolicy*> OpenerList;

    /// constructor
    /**
       Besides those file formats which are supported directly by
       libsndfile, IOAudioFileOpener is able to handle MP3, 
       OGG/Vorbis, MP3+ and LPAC. The corresponding openers are built in.
       Use this constructor to insert additional user defined openers. 
     */
    IOAudioFileOpener
    (
	OpenerList& userOpeners ///< list of user defined openers
    )
    {
	openers = userOpeners;
	openers.push_back(&mp3Opener);
	openers.push_back(&oggOpener);
	openers.push_back(&mppOpener);
	openers.push_back(&lpacOpener);
	openers.push_back(&fileOpener);
    }

    /// default constructor
    /**
       Besides those file formats which are supported directly by
       libsndfile, IOAudioFileOpener is able to handle MP3, 
       OGG/Vorbis, MP3+ and LPAC. The corresponding openers are built in.
     */
    IOAudioFileOpener()
    {
	openers.push_back(&mp3Opener);
	openers.push_back(&oggOpener);
	openers.push_back(&mppOpener);
	openers.push_back(&lpacOpener);
	openers.push_back(&fileOpener);
    }

#if 0
    /// get matching open policy
    /**
       getPolicy() returns the first matching open policy.
     */
    const IOAudioFileOpenPolicy& getPolicy(const std::string& filename) const
    {
	// get extension:
	std::string extension;
	std::string::size_type pos = filename.find_last_of(".");
	if (std::string::npos != pos) 
	    extension = filename.substr(pos+1);

	for (OpenerList::const_iterator cv = openers.begin();
	     cv != openers.end(); ++cv)
	{
	    if ((*cv)->matchesExtension(extension))
		return *(*cv);
	}
	return this->fileOpener;
    }
#endif

    /// open file
    /**
       openFile() uses the first matching opener to open
       the specified file. As a last resort the normal file
       opener is used to open the file directly.
       \return pointer to open file
       \return NULL in case of error
     */
    FILE* openFile
    (
	const std::string& filename, ///< name of file
	std::string& ConverterName   ///< output: set to name of conversion program used to open file
    ) const
    {
	// get extension:
	std::string extension;
	std::string::size_type pos = filename.find_last_of(".");
	if (std::string::npos != pos) 
	    extension = filename.substr(pos+1);

	for (OpenerList::const_iterator cv = openers.begin();
	     cv != openers.end(); ++cv)
	{
	    if ((*cv)->matchesExtension(extension))
		return (*cv)->openFile(filename, ConverterName);
	}
	return 0;
    }

 private:
    OpenerList openers;
    IOAudioFileOpenPolicyMP3  mp3Opener;
    IOAudioFileOpenPolicyOGG  oggOpener;
    IOAudioFileOpenPolicyMPP  mppOpener;
    IOAudioFileOpenPolicyLPAC lpacOpener;
    IOAudioFileOpenPolicyFile fileOpener;
};

//\@} // end of Audio File 'Open' Policy classes
//----------------------------------------------------------------------------------



//----------------------------------------------------------------------------------
/** \defgroup Audio File 'Data Read' Policy classes */
//\@{

/// empty template policy for reading any SampleType from audio file
/**
   This generic template defines the interface to all usable policies.
   It is empty to force compilation errors if someone uses this policy
   with a type which is not supplied as partially specialized template.

   This policy supplies the following functions:
   - int readBuf(SNDFILE* sfp, SampleType* data, int items)

   Andrei Alexandrescu's book "Modern C++ Design" explains this
   type of template policies in chapter 1 (Policy-Based Class Design).
 */
template <typename SampleType> struct IOAudioFileReadPolicy
{
};

/// special policy for reading int32 data from audio file
template <> struct IOAudioFileReadPolicy<int32>
{
    /// read audio data into 32 bit integer buffer
    /**
       \param sfp opaque pointer to libsndfile handle
       \param data 32 bit integer buffer
       \param items number of 32 bit integers in 'data' buffer
       \return number of 32 bit integers read from file
       \return 0 in case of EOF
     */
    static int readBuf(SNDFILE* sfp, int32* data, int items)
	{ return sf_read_int(sfp, data, items); }
};

/// special policy for reading int16 data from audio file
template <> struct IOAudioFileReadPolicy<int16>
{
    /// read audio data into 16 bit integer buffer
    /**
       \param sfp opaque pointer to libsndfile handle
       \param data 16 bit integer buffer
       \param items number of 16 bit integers in 'data' buffer
       \return number of 16 bit integers read from file
       \return 0 in case of EOF
     */
    static int readBuf(SNDFILE* sfp, int16* data, int items)
	{ return sf_read_short(sfp, data, items); }
};

/// special policy for reading 32bit float data from audio file
template <> struct IOAudioFileReadPolicy<float>
{
    /// read audio data into 32 bit float buffer
    /**
       \param sfp opaque pointer to libsndfile handle
       \param data 32 bit float buffer
       \param items number of 32 bit floats in 'data' buffer
       \return number of 32 bit floats read from file
       \return 0 in case of EOF
     */
    static int readBuf(SNDFILE* sfp, float* data, int items)
	{ return sf_read_float(sfp, data, items); }
};

/// special policy for reading 64bit double data from audio file
template <> struct IOAudioFileReadPolicy<double>
{
    /// read audio data into 64 bit double buffer
    /**
       \param sfp opaque pointer to libsndfile handle
       \param data 64 bit double buffer
       \param items number of 64 bit doubles in 'data' buffer
       \return number of 64 bit doubles read from file
       \return 0 in case of EOF
     */
    static int readBuf(SNDFILE* sfp, double* data, int items)
	{ return sf_read_double(sfp, data, items); }
};

//\@} // end of Audio File 'Data Read' Policy classes
//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------
// Error classes
//----------------------------------------------------------------------------------

//----------------------------------------------------------------------------------
/// error class for libsndfile errors
class IOAudioFileError: public ErrorBasic
{
 public:
    /// constructor: extract error fom libsndfile handle  
    IOAudioFileError(SNDFILE* sfp) : ErrorBasic(sf_strerror(sfp)) {}
    /// constructor: extract error from error code
    IOAudioFileError(int errcode) : ErrorBasic(sf_error_number(errcode)) {}
};
//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------
// Utility classes
//----------------------------------------------------------------------------------

//----------------------------------------------------------------------------------
/// utility functions to interpret SF_INFO struct
class IOAudioFileInfo
{
 public:
    /// return audio file information from SF_INFO struct as string
    static std::string getSfInfo(const SF_INFO& info);

    /// return format information from SF_INFO struct member 'format' as string
    static std::string getSfInfoFormat(int format);

    /// get minimum number of sample bits of the audio file
    /**
       \return positive number: number of integer bits
       \return negative number: number of floating point integer bits
       \return 0: number of bits is unknown
     */
    static int getSampleBits(const SF_INFO& info);
};
//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------
// User classes
//----------------------------------------------------------------------------------

//----------------------------------------------------------------------------------
/// C++ interface to library libsndfile
template <typename SampleType> class IOAudioFile
{
 public:
    //-------------------
    // con-/destructors:
    //-------------------
    
    /// constructor for input audio file including compressed formats
    IOAudioFile
    (
	const std::string& name, ///< name of output file
	IOAudioFileOpener& afo, ///< audio file opener
	int verbose = 0		 ///< verbose level
    )
	: fileName(name), verbose(verbose), fp(0), fd(-1), sfp(0)
    {
	memset(&this->sfi, 0, sizeof(this->sfi));  // init SF_INFO

	// use supplied file opener to open the file
	std::string opener;
	this->fp = afo.openFile(fileName, opener);
	if (!this->fp) 
	{
	    if (opener.empty()) throw ErrorSys("cannot open file \"" + this->fileName + "\"");
	    throw ErrorSys("cannot call "+  opener + " \"" + this->fileName + "\"");
	}

	this->attachToHandle();
    }


    /// constructor for input audio file 
    IOAudioFile
    (
	const std::string& name, ///< name of output file
	int verbose = 0		 ///< verbose level
    )
	: fileName(name), verbose(verbose), fp(0), fd(-1), sfp(0)
    {
	memset(&this->sfi, 0, sizeof(this->sfi));  // init SF_INFO

	// use simple file opener to open the file
	IOAudioFileOpenPolicyFile afo;
	this->fp = afo.openFile(fileName);
	if (!this->fp) 
	    throw ErrorSys("cannot open file \"" + this->fileName + "\"");

	this->attachToHandle();
    }

    /// destructor
    virtual ~IOAudioFile()
    {
	// close handle:
	if (this->sfp) sf_close(this->sfp);
	// close open file:
	if (this->fp && (this->fp != stdout)) 
	    (void)fclose(this->fp);
    }

    //------------------
    // access functions
    //------------------
    
    /// get number of samples
    int getNumSamples() const { return sfi.frames; }

    /// get number of channels
    int getNumChannels() const { return sfi.channels; }

    /// get sample rate in Hz
    int getSampleRate() const { return sfi.samplerate; }

    /// get runtime in seconds
    double getTime() const
	{ return ((double)sfi.frames / (double)sfi.samplerate); }

    /// get runtime of specified frames in seconds
    double getTime(uint32 frameNum) const
	{ return ((double)frameNum / (double)sfi.samplerate); }

    //-----------------------
    // read/write functions:
    //-----------------------

    /// read one frame (1 sample per channel) from audio file
    /** <tt>readFrame()</tt> reads one sample of each channel into the
        <tt>frame</tt> buffer. Therefore the <tt>frame</tt> buffer must 
	be big enough to contain <tt>getNumChannels()</tt> samples. 
     */
    int32 readFrame
    (
	SampleType* frame ///< pointer to frame buffer
    )
    { return this->cache.readFrame(this->sfp, frame, this->sfi.channels); }

    /// read one frame (1 sample per channel) from audio file
    /** <tt>readFrame()</tt> reads one sample of each channel into the
        <tt>frame</tt> vector. The frame vector is not cleared in advance.
     */
    int32 readFrame
    (
	std::vector<SampleType>& frame ///< frame vector
    )
    { return this->cache.readFrame(this->sfp, frame, this->sfi.channels); }

    //-----------------
    // info functions:
    //-----------------

    /// get minimum number of sample bits of the audio file
    /**
       \return positive number: number of integer bits
       \return negative number: number of floating point integer bits
       \return 0: number of bits is unknown
     */
    int getSampleBits() const
	{ return IOAudioFileInfo::getSampleBits(this->sfi); }

    /// print information about the audio file
    std::string getInfo() const
	{ return this->fileName + ":\n" + IOAudioFileInfo::getSfInfo(this->sfi); }

 protected:
    void attachToHandle()
    {
	// get file descriptor from file pointer
	this->fd = fileno(this->fp);
	if (-1 == this->fd)
	{
	    if (this->fp != stdout) (void)fclose(this->fp);
	    throw ErrorSys(this->fileName);
	}

	// attach the open file to libsndfile:
	this->sfi.format = 0;
	this->sfp = sf_open_fd (this->fd, SFM_READ, &this->sfi, 0);
	if (!this->sfp) 
	{
	    if (this->fp != stdout) (void)fclose(this->fp);
	    throw ErrorBasic("cannot open " + this->fileName);
	}

	if (1 < this->verbose)
	    cerr << IOAudioFileInfo::getSfInfo(this->sfi);
    }

    const std::string fileName;
    int 	verbose;

 private:
    class SfCache
    {
    public:
	SfCache(int32 cacheSize = 64*1024) : data(0), size(cacheSize), num(0), pos(0) {}
	~SfCache() { delete[] data; }

	/// read one frame
	int readFrame(SNDFILE* sfp, SampleType* frame, int channels)
	{
	    for(int c = 0; c < channels; ++c)
	    {
		int ret = readSample(sfp, frame + c);
		if (ret < 1) return ret;
	    }
	    return 1;
	}

	/// read one frame
	int readFrame(SNDFILE* sfp, std::vector<SampleType>& frame, int channels)
	{
	    for(int c = 0; c < channels; ++c)
	    {
		SampleType value;
		int ret = readSample(sfp, &value);
		if (ret < 1) return ret;
		frame.push_back(value);
	    }
	    return 1;
	}
    private:
	/// read one sample
	int readSample(SNDFILE* sfp, SampleType* value)
	{
	    if (!data) 
	    {
		// create cache buffer
		data = new SampleType[size];
		if (!data) throw ErrorSys(ENOMEM);
	    }
	    if (pos == num) 
	    {
		// cache buffer completely read: fill it with next data
		num = readPolicy.readBuf(sfp, data, size);
		if (num < 1) return num; // EOF?
		pos = 0; // set to begin of buffer
	    }
	    *value = data[pos++];
	    return 1;
	}

	IOAudioFileReadPolicy<SampleType> readPolicy;
	SampleType* data;
	int32 size;
	int32 num;
	int32 pos;
    };

    FILE*	fp;             // file pointer
    int         fd;             // file descriptor
    SNDFILE*    sfp;            // opaque handle to libsndfile
    SF_INFO     sfi;            // info struct of libsndfile
    SfCache     cache;          // file cache
}; // class IOAudioFile
//----------------------------------------------------------------------------------


END_NAMESPACE

#endif // !AH_IO_AUDIO_FILE_H
