//---------------------------------------------------------------------------
// Copyright 1998 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/IO/CFormat.h,v $
// $Id: CFormat.h,v 1.4 2004/03/21 06:05:39 ah Exp $
//
// Description: C-style printf compatible ostream applicator
//
//---------------------------------------------------------------------------

#ifndef AH_IO_CFORMAT_H		// prevent multiple includes
#define AH_IO_CFORMAT_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
USING_NAMESPACE(AH)

#include <string>
#include <iostream>
USING_NAMESPACE(std)

// --- types and classes ---

BEGIN_NAMESPACE(AH)

/// C-style printf compatible ostream applicator
/**
   <tt>IOCFormat</tt> is a C++ <tt>ostream</tt> manipulator which takes a 
   <tt>printf(3S)</tt> format string as constructor argument and sets the
   <tt>ostream</tt> flags accordingly when its public member function 
   <tt>manip()</tt> is applied on an <tt>ostream</tt>. 

   For convenience the <tt>ostream</tt> output operator is overloaded, so 
   just putting an <tt>IOCFormat</tt> object (even an temporary one) into 
   an output stream will be sufficient, see the following example:

   \verbatim
   #include <AH/IO/CFormat.h>

   cout << IOCFormat("%5.2f") << 3.141592653 << endl;
   \endverbatim

   A local <tt>typedef</tt> makes this even shorter to write and understand:

   \verbatim
   #include <AH/IO/CFormat.h>

   typedef IOCFormat fmt;
   cout << fmt("%5.2f") << 3.141592653 << endl;
   \endverbatim

   If the same format is used often it is recommended to define it once 
   and use it multiply, as then the format conversion is done only once 
   (at construction time):

   \verbatim
   #include <AH/IO/CFormat.h>

   IOCFormat fmt("%10.2f");
   for (int i = 0; i < 1000; i++)
   cout << fmt << (3.141592653 + i) << endl;
   \endverbatim

   <tt>IOCFormat</tt> supports the following formats (refer to the 
   <tt>printf(3S)</tt> man pages for explanations):
   - flags: -, +, \#, 0 (not on strings)
   - width: a number specifying the field width
   - precision: a decimal point followed by a number specifying the precision
   - size: h, l, ll, and L are accepted but have no effect
   - type: d, x, X, o, f, e, E, s

   Please include  {\em AH/IO/CFormat.h}
 */
class IOCFormat
{
 public:
    /// constructor
    IOCFormat(const string& format);
    /// default constructor
    IOCFormat(const char* format = 0);

    /// manipulate ostream
    ostream& manip(ostream& out) const;

 private:
    void init(const char* format);

    char		fill;
    int			precision;
    int			width;
    ios::fmtflags	base;
    ios::fmtflags	floating;
    ios::fmtflags	adjust;
    ios::fmtflags	showpos;
    bool		showbase;
    bool		setupper;
    bool		setprecision;
    bool		hash;
    bool		error;
};

/// allow to insert IOCFormat objects directly into an ostream
inline ostream& operator<< (ostream& out, const IOCFormat& fmt)
{
    return fmt.manip(out);
}

END_NAMESPACE

#endif // !AH_IO_CFORMAT_H
