//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/MpegFrameHeader.c,v $
// $Id: MpegFrameHeader.c,v 1.5 2004/04/20 22:26:22 ah Exp $ 
//
// Description: Implementation of class AH::IOMpegFrameHeader
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: MpegFrameHeader.c,v 1.5 2004/04/20 22:26:22 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
#include <AH/IO/Mpeg/FrameHeader.h>
USING_NAMESPACE(AH)

#include <string>
#include <cstdio>
USING_NAMESPACE(std)

BEGIN_NAMESPACE(AH)

// --- local constants ---

// table of bitrates in kB/s for MPEG-1 and MPEG-2, Layer 1-3
const int bitrates[16][6] =
{
    // -- MPEG1 ---  -- MPEG-2 ---
    // L1   L2   L3   L1   L2   L3
    {   0,   0,   0,   0,   0,   0 },
    {  32,  32,  32,  32,  32,   8 },
    {  64,  48,  40,  64,  48,  16 },
    {  96,  56,  48,  96,  56,  24 },
    { 128,  64,  56, 128,  64,  32 },
    { 160,  80,  64, 160,  80,  40 },
    { 192,  96,  80, 192,  96,  48 },
    { 224, 112,  96, 224, 112,  56 },
    { 256, 128, 112, 256, 128,  64 },
    { 288, 160, 128, 288, 160,  80 },
    { 320, 192, 160, 320, 192,  96 },
    { 352, 224, 192, 352, 224, 112 },
    { 384, 256, 224, 384, 256, 128 },
    { 416, 320, 256, 416, 320, 144 },
    { 448, 384, 320, 448, 384, 160 },
    {  -1,  -1,  -1,  -1,  -1,  -1 }
};

// table of samplerates in samples/s for MPEG-1, MPEG-2 and MPEG-2.5
const int samplerates[4][3] =
{
    // MPEG1  MPEG-2 MPEG2.5
    {  44100, 22050, 11025 },
    {  48000, 24000, 12000 },
    {  32000, 16000,  8000 },
    {     -1,    -1,    -1 }
};

// --- member functions ---


// constructor
IOMpegFrameHeader::IOMpegFrameHeader(const uint8* frameData)
{
    if (frameData) memcpy(bytes, frameData, 4);
    else           memset(bytes, 0, 4);
}


// return true if frame is valid
bool IOMpegFrameHeader::validHeader() const
{
    // check for first 11 bits set:
    if (!isFirstByte(bytes[0])) return false;
    if (!isSecondByte(bytes[0])) return false;

    // check valid version and rates:
    if (-1 == getMpegVersion()) return false;
    if (-1 == getBitRate()) return false;
    if (-1 == getSampleRate()) return false;

    return true;
}


// get version of MPEG: 1=MPEG1, 2=MPEG2, 3=MPEG2.5, -1=invalid
int IOMpegFrameHeader::getMpegVersion() const
{
    int code = (bytes[1] >> 3) & 0x3;
    switch (code)
    {
    case 0: return 3;
    default: // keep compiler quiet
    case 1: return -1;
    case 2: return 2;
    case 3: return 1;
    }
}


// get layer: 1/2/3=Layer 1/2/3, -1=invalid
int IOMpegFrameHeader::getLayer() const
{
    int code = (bytes[1] >> 1) & 0x3;
    switch(code)
    {
    default: // keep compiler quiet
    case 0: return -1;
    case 1: return 3;
    case 2: return 2;
    case 3: return 1;
    }
}


// get bit rate in kBit/sec: 0=variable, -1=invalid
int IOMpegFrameHeader::getBitRate() const
{
    // get MPEG layer:
    int mpeg = getMpegVersion();
    int idx;
    switch (mpeg)
    {
    default: // fall through
    case -1: return -1;
    case  1: idx = -1; break;
    case  2: // fall through
    case  3: idx = 2; break;
    }

    int layer = getLayer();
    if (-1 == layer) return -1;

    // calculate index into bitrate table:
    idx += layer;
    if ((0 > idx) || (6 <= idx)) return -1;

    int code = (bytes[2] >> 4) & 0xf;
    return bitrates[code][idx];
}


// get sample rate in kSamples/sec: -1=invalid
int IOMpegFrameHeader::getSampleRate() const
{
    // get MPEG layer:
    int mpeg = getMpegVersion();

    // calculate index into bitrate table:
    int idx;
    switch (mpeg)
    {
    default: // fall through
    case -1: return -1;
    case  1: // fall through
    case  2: // fall through
    case  3: idx = mpeg-1; break;
    }

    int code = (bytes[2] >> 2) & 0x3;
    return samplerates[code][idx];
}


// return true if frame contains CRC checksum
bool IOMpegFrameHeader::hasCRC() const
{
    int code = bytes[1] & 0x1;
    if (code) return false;
    return true;
}


// get padding byte
int IOMpegFrameHeader::getPadding() const
{
    int code = (bytes[2] >> 1) & 0x1;
    return code;
}


int IOMpegFrameHeader::getFrameSize() const
{
    // get MPEG layer:
    int mpeg = getMpegVersion();
    if (-1 == mpeg) return -1;
    if (3 == mpeg) mpeg = 2;

    int layer = getLayer();
    if (-1 == layer) return -1;

    // get factor:
    int factor = 0;
    switch (mpeg)
    {
    case 1:
	switch (layer)
	{
	case 1: factor = 48000; break;
	case 2: // fall through
	case 3: factor = 144000; break;
	}
	break;
    case 2:
	switch (layer)
	{
	case 1: factor = 24000; break;
	case 2: // fall through
	case 3: factor = 72000; break;
	}
	break;
    }

    int bitrate = getBitRate();
    if (-1 == bitrate) return -1;
    int samplerate = getSampleRate();
    if (-1 == samplerate) return -1;

    int length = (factor * bitrate) / samplerate + getPadding();
    return length;
}


// get CRC data size
int IOMpegFrameHeader::getCrcSize() const
{
    if (hasCRC()) return 2;
    return 0;
}


// get audio data size
int IOMpegFrameHeader::getAudioSize() const
{
    int length = getFrameSize();
    if (-1 == length) return -1;

    length -= 4 + getCrcSize();
    if (0 > length) return -1;
    return length;
}


// get audio time
double IOMpegFrameHeader::getFrameTime() const
{
    int br = getBitRate();
    if (-1 == br) return -1;

    int as = getFrameSize();
    if (-1 == as) return -1;

    double timval = (double)(as * 8)/(double)br;
    return timval;
}


/// get info string
string IOMpegFrameHeader::getInfo() const
{
    char buf[80];
    string info("MPEG Frame Info:\n");

    (void)sprintf(buf, "raw header: %#04x %#04x %#04x %#04x\n",
		  bytes[0], bytes[1], bytes[2], bytes[3]);
    info += buf;

    // get MPEG Version and Layer:
    info += "Version: ";
    switch(getMpegVersion())
    {
    default: info += "unknown"; break;
    case 1: info += "MPEG-1"; break;
    case 2: info += "MPEG-2"; break;
    case 3: info += "MPEG-2.5"; break;
    }
    info += " Layer ";
    switch(getLayer())
    {
    default: info += "unknown"; break;
    case 1: info += "1"; break;
    case 2: info += "2"; break;
    case 3: info += "3"; break;
    }
    info += "\n";

    // get audio info:
    info += "Samplerate: ";
    int sr = getSampleRate();
    if (-1 == sr) info += "unknown";
    else 
    {
	(void)sprintf(buf, "%d Hz", sr);
	info += buf;
    }
    info += "\n";

    info += "Bitrate: ";
    int br = getBitRate();
    if (-1 == br) info += "unknown";
    else if (0 == br) info += "variable";
    else 
    {
	(void)sprintf(buf, "%d kB/sec", br);
	info += buf;
    }
    info += "\n";
   
    info += "Framelength: ";
    int fl = getFrameSize();
    if (-1 == fl) info += "unknown";
    else 
    {
	(void)sprintf(buf, "%d bytes", fl);
	info += buf;
    }
    info += "\n";
   
    // get time:
    info += "Frametime: ";
    double at = getFrameTime();
    if (0 > at)  info += "unknown";
    else
    {
	(void)sprintf(buf, "%7.3f ms", at);
	info += buf;
    }
    info += "\n";
    return info;
}


// return true if byte is first byte of a MPEG frame header
bool IOMpegFrameHeader::isFirstByte(uint8 byte)
{
    if (0xff != byte) return false;
    return true;
}


// return true if byte is second byte of a MPEG frame header
bool IOMpegFrameHeader::isSecondByte(uint8 byte)
{
    if (0xe0 != (byte & 0xe0)) return false;
    return true;
}

END_NAMESPACE

// --- end of file ---
